# keyPoint Matching
from .keypoint_matching.kaze import KAZE
from .keypoint_matching.orb import ORB
from .keypoint_matching.sift import SIFT, RootSIFT
from .keypoint_matching.surf import SURF
from .keypoint_matching.brief import BRIEF

# CUDA keyPoint Matching
from .keypoint_matching.cuda.surf import CUDA_SURF
from .keypoint_matching.cuda.orb import CUDA_ORB


# Template Matching
from .template_matching.matchTemplate import MatchTemplate

# CUDA Template Matching
from .template_matching.cuda.matchTemplate import CudaMatchTemplate


# api
from .findit import Findit




# -*- coding: utf-8 -*-
from typing import Union

import cv2
import numpy as np

from image_registration.keypoint_matching.kaze import KAZE
from image_registration.exceptions import (CreateExtractorError, NoEnoughPointsError)


class SIFT(KAZE):
    METHOD_NAME = "SIFT"

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True,
                 nfeature: int = 0, nOctaveLayers: int = 3, contrastThreshold: float = 0.04, edgeThreshold: int = 10,
                 sigma: Union[int, float] = 1.6, descriptorType: int = cv2.CV_32F):
        super(SIFT, self).__init__(threshold, rgb)
        # 初始化参数
        self.extractor_parameters = dict(
            nfeatures=nfeature, nOctaveLayers=nOctaveLayers, contrastThreshold=contrastThreshold,
            edgeThreshold=edgeThreshold, sigma=sigma, descriptorType=descriptorType
        )
        # 创建SIFT实例
        try:
            self.detector = cv2.SIFT_create(**self.extractor_parameters)
        except Exception:
            raise CreateExtractorError('create sift extractor error')


class RootSIFT(SIFT):
    METHOD_NAME = 'RootSIFT'

    def get_keypoints_and_descriptors(self, image):
        keypoints, descriptors = self.detector.detectAndCompute(image, None)
        keypoints, descriptors = self.rootSIFT_compute(image, keypoints)

        if len(keypoints) < 2:
            raise NoEnoughPointsError('{} detect not enough feature points in input images'.format(self.METHOD_NAME))
        return keypoints, descriptors

    def rootSIFT_compute(self, image, kps, eps=1e-7):
        keypoints, descriptors = self.detector.compute(image, kps)

        if len(keypoints) == 0:
            return [], None

        descriptors /= (descriptors.sum(axis=1, keepdims=True) + eps)
        descriptors = np.sqrt(descriptors)

        return keypoints, descriptors



# -*- coding: utf-8 -*-
import cv2
import numpy as np
from baseImage import Image
from image_registration.keypoint_matching.kaze import KAZE
from typing import Tuple, List, Union
from image_registration.exceptions import (CreateExtractorError, NoEnoughPointsError,
                                           CudaSurfInputImageError)


class CUDA_SURF(KAZE):
    METHOD_NAME = "CUDA_SURF"
    # 方向不变性:True检测/False不检测
    UPRIGHT = True
    # 检测器仅保留其hessian大于hessianThreshold的要素,值越大,获得的关键点就越少
    HESSIAN_THRESHOLD = 400

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True,
                 _hessianThreshold=400, _nOctaves=4, _nOctaveLayers=2, _extended=True, _upright=False):
        super(CUDA_SURF, self).__init__(threshold, rgb)
        # 初始化参数
        self.extractor_parameters = dict(
            _hessianThreshold=_hessianThreshold, _nOctaves=_nOctaves, _nOctaveLayers=_nOctaveLayers,
            _extended=_extended, _upright=_upright,
        )
        try:
            self.detector = cv2.cuda.SURF_CUDA_create(**self.extractor_parameters)
        except Exception:
            raise CreateExtractorError('create cuda_surf extractor error')

    def create_matcher(self):
        """
        创建特征点匹配器
        :return: Brute-force descriptor matcher
        """
        # matcher = cv2.cuda.DescriptorMatcher_createBFMatcher(cv2.NORM_L2)
        matcher = cv2.cuda_DescriptorMatcher.createBFMatcher(cv2.NORM_L2)
        return matcher

    def check_image_input(self, im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                          im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes]) -> Tuple[Image, Image]:
        """
        检测输入的图像数据是否正确,并且转换为Gpu模式
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :return im_source, im_search
        """
        if not isinstance(im_source, Image):
            im_source = Image(im_source)
        if not isinstance(im_search, Image):
            im_search = Image(im_search)

        self._check_image_size(im_source)
        self._check_image_size(im_search)

        im_source.transform_gpu()
        im_search.transform_gpu()

        return im_source, im_search

    def _check_image_size(self, image: Image):
        """
            SURF匹配特征点时,无法处理长宽太小的图片
            https://stackoverflow.com/questions/42492060/surf-cuda-error-while-computing-descriptors-and-keypoints
            https://github.com/opencv/opencv_contrib/blob/master/modules/xfeatures2d/src/surf.cuda.cpp#L151
        """
        # (9 + 6 * 0) << nOctaves-1

        def calc_size(octave, layer):
            HAAR_SIZE0 = 9
            HAAR_SIZE_INC = 6
            return (HAAR_SIZE0 + HAAR_SIZE_INC * layer) << octave

        min_size = int(calc_size(self.detector.nOctaves - 1, 0))
        layer_height = image.size[0] >> (self.detector.nOctaves - 1)
        layer_width = image.size[1] >> (self.detector.nOctaves - 1)
        min_margin = ((calc_size((self.detector.nOctaves - 1), 2) >> 1) >> (self.detector.nOctaves - 1)) + 1

        if image.size[0] - min_size < 0 or image.size[1] - min_size < 0:
            raise CudaSurfInputImageError('The image size({width}x{height}) does not conform to SURF_CUDA standard'.
                                          format(width=image.size[1], height=image.size[0]))
        if layer_height - 2 * min_margin < 0 or layer_width - 2 * min_margin < 0:
            raise CudaSurfInputImageError('The image size({width}x{height}) does not conform to SURF_CUDA standard'.
                                          format(width=image.size[1], height=image.size[0]))

    def get_rect_from_good_matches(self, im_source: Image, im_search: Image,
                                   kp_sch, des_sch: cv2.cuda_GpuMat,
                                   kp_src, des_src: cv2.cuda_GpuMat):
        # TODO: 增加kp_sch,des_sch的类型提示,以及范函数返回值的类型提示
        matches = self.match_keypoints(des_sch=des_sch, des_src=des_src)
        good = self.get_good_in_matches(matches)

        kp_sch = self.detector.downloadKeypoints(kp_sch)
        kp_src = self.detector.downloadKeypoints(kp_src)

        rect = self.extract_good_points(im_source, im_search, kp_sch, kp_src, good)
        return rect, matches, good

    def match_keypoints(self, des_sch: cv2.cuda_GpuMat, des_src: cv2.cuda_GpuMat) -> List[List[cv2.DMatch]]:
        """Match descriptors (特征值匹配)."""
        # 匹配两个图片中的特征点集，k=2表示每个特征点取出2个最匹配的对应点:
        matches = self.matcher.knnMatch(des_sch, des_src, 2)
        return matches

    def get_keypoints_and_descriptors(self, image: cv2.cuda_GpuMat) -> Tuple[List[cv2.KeyPoint], cv2.cuda_GpuMat]:
        """获取图像特征点和描述符."""
        keypoints, descriptors = self.detector.detectWithDescriptors(image, None)

        if keypoints.size()[0] < 2:
            raise NoEnoughPointsError('{} detect not enough feature points in input images'.format(self.METHOD_NAME))
        return keypoints, descriptors



#! usr/bin/python
# -*- coding:utf-8 -*-
""" opencv matchTemplate"""
import cv2
import numpy as np
import time
from image_registration.exceptions import (MatchResultError)
from image_registration.utils import generate_result, print_run_time
from baseImage import Image, Rect
from typing import Union, Tuple


class MatchTemplate(object):
    METHOD_NAME = "tpl"

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True):
        """
        初始化
        :param threshold: 识别阈值(0~1)
        :param rgb: 是否使用rgb通道进行校验
        :return: None
        """
        self.threshold = threshold
        self.rgb = rgb

    @print_run_time()
    def find_best_result(self, im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         threshold: Union[int, float] = None, rgb: bool = True):
        """
        模板匹配
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :param threshold: 识别阈值(0~1)
        :param rgb: 是否使用rgb通道进行校验
        :return: None or Rect
        """
        im_source, im_search = self.check_detection_input(im_source, im_search)
        result = self._get_template_result_matrix(im_source, im_search)

        # 找到最佳匹配项
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
        h, w = im_search.size
        # 求可信度
        crop_rect = Rect(max_loc[0], max_loc[1], w, h)
        confidence = self.cal_confidence(im_source, im_search, crop_rect, max_val, rgb)
        # 如果可信度小于threshold,则返回None

        if confidence < (threshold or self.threshold):
            return None
        x, y = max_loc
        rect = Rect(x=x, y=y, width=w, height=h)
        return generate_result(rect, confidence)

    def find_all_results(self, im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         threshold: Union[int, float] = None, max_count: int = 10, rgb: bool = True):
        """
        模板匹配
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :param threshold: 识别阈值(0~1)
        :param max_count: 最多匹配数量
        :param rgb: 是否使用rgb通道进行校验
        :return: None or Rect
        """
        im_source, im_search = self.check_detection_input(im_source, im_search)
        # 模板匹配取得矩阵
        res = self._get_template_result_matrix(im_source, im_search)
        result = []
        h, w = im_search.size
        while True:
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
            img_crop = im_source.crop_image(Rect(max_loc[0], max_loc[1], w, h))
            confidence = self._get_confidence_from_matrix(img_crop, im_search, max_val=max_val, rgb=rgb)
            x, y = max_loc
            rect = Rect(x, y, w, h)
            if (confidence < (threshold or self.threshold)) or len(result) >= max_count:
                break
            result.append(generate_result(rect, confidence))
            cv2.rectangle(res, (int(max_loc[0] - w / 2), int(max_loc[1] - h / 2)),
                          (int(max_loc[0] + w / 2), int(max_loc[1] + h / 2)), (0, 0, 0), -1)
        return result if result else None

    @staticmethod
    def _get_template_result_matrix(im_source: Image, im_search: Image) -> np.ndarray:
        """求取模板匹配的结果矩阵."""
        s_gray, i_gray = im_search.rgb_2_gray(), im_source.rgb_2_gray()
        return cv2.matchTemplate(i_gray, s_gray, cv2.TM_CCOEFF_NORMED)

    @staticmethod
    def check_detection_input(im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                              im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes]) -> Tuple[Image, Image]:
        """
        检测输入的图像数据是否正确
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :return: im_source, im_search
        """
        if not isinstance(im_source, Image):
            im_source = Image(im_source)
        if not isinstance(im_search, Image):
            im_search = Image(im_search)

        im_source.transform_cpu()
        im_search.transform_cpu()
        return im_source, im_search

    @staticmethod
    def cal_rgb_confidence(im_source: Image, im_search: Image):
        """
        计算两张图片图片rgb三通道的置信度
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :return: 最小置信度
        """
        img_src_rgb, img_sch_rgb = im_source.imread(), im_search.imread()
        img_sch_rgb = cv2.copyMakeBorder(img_sch_rgb, 10, 10, 10, 10, cv2.BORDER_REPLICATE)
        # 转HSV强化颜色的影响
        img_src_rgb = cv2.cvtColor(img_src_rgb, cv2.COLOR_BGR2HSV)
        img_sch_rgb = cv2.cvtColor(img_sch_rgb, cv2.COLOR_BGR2HSV)
        src_bgr, sch_bgr = cv2.split(img_src_rgb), cv2.split(img_sch_rgb)
        # 计算BGR三通道的confidence，存入bgr_confidence:
        bgr_confidence = [0, 0, 0]
        for i in range(3):
            res_temp = cv2.matchTemplate(src_bgr[i], sch_bgr[i], cv2.TM_CCOEFF_NORMED)
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res_temp)
            bgr_confidence[i] = max_val
        return min(bgr_confidence)

    @staticmethod
    def cal_ccoeff_confidence(im_source: Image, im_search: Image):
        """
        使用CCOEFF方法模板匹配图像
        """
        img_src_gray, img_sch_gray = im_source.rgb_2_gray(), im_search.rgb_2_gray()
        # 扩展置信度计算区域
        img_sch_gray = cv2.copyMakeBorder(img_sch_gray, 10, 10, 10, 10, cv2.BORDER_REPLICATE)

        res = cv2.matchTemplate(img_src_gray, img_sch_gray, cv2.TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)

        return max_val

    def cal_confidence(self, im_source: Image, im_search: Image, crop_rect: Rect, max_val, rgb) -> Union[int, float]:
        """
        将截图和识别结果缩放到大小一致,并计算可信度
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :param crop_rect: 需要在im_source截取的区域
        :param rgb: 是否使用rgb通道进行校验
        :param max_val: matchTemplate得到的最大值
        :raise MatchResultError: crop_rect范围超出了im_source边界
        :return: 返回可信度(0~1)
        """
        try:
            target_img = im_source.crop_image(crop_rect)
        except OverflowError:
            raise MatchResultError(f"Target area({crop_rect}) out of screen{im_source.size}")
        confidence = self._get_confidence_from_matrix(target_img, im_search, max_val, rgb)
        return confidence

    @staticmethod
    def _get_confidence_from_matrix(img_crop, im_search, max_val, rgb):
        """根据结果矩阵求出confidence."""
        # 求取可信度:
        if rgb:
            # 如果有颜色校验,对目标区域进行BGR三通道校验:
            confidence = MatchTemplate.cal_rgb_confidence(img_crop, im_search)
        else:
            confidence = max_val
        return confidence



import cv2
from image_registration import (ORB, RootSIFT, SIFT, SURF, KAZE, BRIEF, MatchTemplate,
                                CUDA_SURF, CUDA_ORB, CudaMatchTemplate)

CVSTRATEGY = [MatchTemplate, ORB, RootSIFT]
if cv2.cuda.getCudaEnabledDeviceCount() > 0:
    CVSTRATEGY = [CudaMatchTemplate, CUDA_ORB, RootSIFT]

CVPARAMS = {
    'CUDA_ORB': [
        dict(nfeatures=60000, scaleFactor=2, nlevels=1, firstLevel=1),
    ]
}


# -*- coding: utf-8 -*-
import cv2
import numpy as np
from typing import Union, Tuple, List
from baseImage import Image
from image_registration.keypoint_matching.kaze import KAZE
from image_registration.exceptions import (CreateExtractorError, CudaOrbDetectorError,
                                           NoEnoughPointsError)


class CUDA_ORB(KAZE):
    METHOD_NAME = 'CUDA_ORB'

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True,
                 nfeatures: int = 50000, scaleFactor: Union[int, float] = 1.2, nlevels: int = 8,
                 edgeThreshold: int = 31, firstLevel: int = 0, WTA_K: int = 2,
                 scoreType: int = cv2.ORB_HARRIS_SCORE, patchSize: int = 31, fastThreshold: int = 20,
                 blurForDescriptor: bool = False):
        super(CUDA_ORB, self).__init__(threshold, rgb)
        # 初始化参数
        self.extractor_parameters = dict(
            nfeatures=nfeatures, scaleFactor=scaleFactor, nlevels=nlevels,
            edgeThreshold=edgeThreshold, firstLevel=firstLevel, WTA_K=WTA_K,
            scoreType=scoreType, patchSize=patchSize, fastThreshold=fastThreshold,
            blurForDescriptor=blurForDescriptor,
        )
        try:
            # 创建ORB实例
            self.detector = cv2.cuda_ORB.create(**self.extractor_parameters)
        except Exception:
            raise CreateExtractorError('create cuda_orb extractor error')

    def check_image_input(self, im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                          im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes]) -> Tuple[Image, Image]:
        """
        检测输入的图像数据是否正确,并且转换为Gpu模式
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :return im_source, im_search
        """
        if not isinstance(im_source, Image):
            im_source = Image(im_source)
        if not isinstance(im_search, Image):
            im_search = Image(im_search)

        im_source.transform_gpu()
        im_search.transform_gpu()
        return im_source, im_search

    def create_matcher(self):
        """
        创建特征点匹配器
        :return: Brute-force descriptor matcher
        """
        matcher = cv2.cuda_DescriptorMatcher.createBFMatcher(cv2.NORM_HAMMING)
        return matcher

    def get_keypoints_and_descriptors(self, image: cv2.cuda_GpuMat) -> Tuple[list, cv2.cuda_GpuMat]:
        # https://github.com/prismai/opencv_contrib/commit/d7d6360fceb5881d596be95b03568d4dcdb7236d#diff-122b9c09d35cd89b7cee1eeb66189e4820f5663bbeda844908f05bf730c93e49
        try:
            keypoints, descriptors = self.detector.detectAndComputeAsync(image, None)
        except cv2.error:
            raise CudaOrbDetectorError('{} detect error, Try adjust detector params'.format(self.METHOD_NAME))
        else:
            keypoints = self.detector.convert(keypoints)

        if len(keypoints) < 2:
            raise NoEnoughPointsError('{} detect not enough feature points in input images'.format(self.METHOD_NAME))
        return keypoints, descriptors

    def match_keypoints(self, des_sch: cv2.cuda_GpuMat, des_src: cv2.cuda_GpuMat) -> List[List[cv2.DMatch]]:
        """
        特征点匹配
        :param des_sch: 图片模板的特征点集
        :param des_src: 待匹配图像的特征点集
        :return: 返回一个列表,包含最匹配的对应点
        """
        # k=2表示每个特征点取出2个最匹配的对应点
        matches = self.matcher.knnMatch(des_sch, des_src, 2)
        return matches

    @staticmethod
    def delect_rect_descriptors(rect, kp, des):
        tl, br = rect.tl, rect.br
        kp = kp.copy()
        des = des.download()

        delect_list = tuple(kp.index(i) for i in kp if tl.x <= i.pt[0] <= br.x and tl.y <= i.pt[1] <= br.y)
        for i in sorted(delect_list, reverse=True):
            kp.pop(i)

        des = np.delete(des, delect_list, axis=0)
        mat = cv2.cuda_GpuMat()
        mat.upload(des)
        return kp, mat

    @staticmethod
    def delect_good_descriptors(good, kp, des):
        kp = kp.copy()
        des = des.download()

        delect_list = [i.trainIdx for i in good]
        for i in sorted(delect_list, reverse=True):
            kp.pop(i)

        des = np.delete(des, delect_list, axis=0)
        mat = cv2.cuda_GpuMat()
        mat.upload(des)
        return kp, mat



# -*- coding: utf-8 -*-
from typing import Union

import cv2
import numpy as np

from image_registration.keypoint_matching.kaze import KAZE
from image_registration.exceptions import (NoEnoughPointsError, NoModuleError)


class BRIEF(KAZE):
    METHOD_NAME = "BRIEF"

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True):
        super(BRIEF, self).__init__(threshold, rgb)
        try:
            # Initiate FAST detector
            self.star = cv2.xfeatures2d.StarDetector_create()
            # Initiate BRIEF extractor
            self.detector = cv2.xfeatures2d.BriefDescriptorExtractor_create()
        except AttributeError:
            raise NoModuleError

    def create_matcher(self) -> cv2.BFMatcher:
        matcher = cv2.BFMatcher_create(cv2.NORM_L1)
        return matcher

    def get_keypoints_and_descriptors(self, image: np.ndarray):
        # find the keypoints with STAR
        kp = self.star.detect(image, None)
        # compute the descriptors with BRIEF
        keypoints, descriptors = self.detector.compute(image, kp)

        if len(keypoints) < 2:
            raise NoEnoughPointsError('{} detect not enough feature points in input images'.format(self.METHOD_NAME))
        return keypoints, descriptors



# -*- coding: utf-8 -*-
import cv2
import numpy as np
from typing import Tuple, List, Union
from image_registration.keypoint_matching.kaze import KAZE
from image_registration.exceptions import (CreateExtractorError, NoModuleError, NoEnoughPointsError)


class ORB(KAZE):
    METHOD_NAME = "ORB"

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True,
                 nfeatures: int = 50000, scaleFactor: Union[int, float] = 1.2, nlevels: int = 8,
                 edgeThreshold: int = 31, firstLevel: int = 0, WTA_K: int = 2,
                 scoreType: int = cv2.ORB_HARRIS_SCORE, patchSize: int = 31, fastThreshold: int = 20):
        super(ORB, self).__init__(threshold, rgb)
        # 初始化参数
        self.extractor_parameters = dict(
            nfeatures=nfeatures, scaleFactor=scaleFactor, nlevels=nlevels,
            edgeThreshold=edgeThreshold, firstLevel=firstLevel, WTA_K=WTA_K,
            scoreType=scoreType, patchSize=patchSize, fastThreshold=fastThreshold,
        )
        try:
            # 创建ORB实例
            self.detector = cv2.ORB_create(**self.extractor_parameters)
        except Exception:
            raise CreateExtractorError('create orb extractor error')
        else:
            try:
                # https://docs.opencv.org/master/d7/d99/classcv_1_1xfeatures2d_1_1BEBLID.html
                # https://github.com/iago-suarez/beblid-opencv-demo
                self.descriptor = cv2.xfeatures2d.BEBLID_create(0.75)
            except AttributeError:
                raise NoModuleError

    def create_matcher(self) -> cv2.DescriptorMatcher_create:
        matcher = cv2.DescriptorMatcher_create(cv2.DescriptorMatcher_BRUTEFORCE_HAMMING)
        return matcher

    def get_good_in_matches(self, matches: list) -> List[cv2.DMatch]:
        """
        特征点过滤
        :param matches: 特征点集
        """
        good = []
        # 出现过matches对中只有1个参数的情况,会导致遍历的时候造成报错
        for v in matches:
            if len(v) == 2:
                if v[0].distance < self.FILTER_RATIO * v[1].distance:
                    good.append(v[0])
        return good

    def get_keypoints_and_descriptors(self, image: np.ndarray) -> Tuple[List[cv2.KeyPoint], np.ndarray]:
        """
        获取图像关键点(keypoints)与描述符(descriptors)
        :param image: 待检测的灰度图像
        :raise NoEnoughPointsError: 检测特征点数量少于2时,弹出异常
        :return: 关键点(keypoints)与描述符(descriptors)
        """
        keypoints = self.detector.detect(image, None)
        keypoints, descriptors = self.descriptor.compute(image, keypoints)

        if len(keypoints) < 2:
            raise NoEnoughPointsError('{} detect not enough feature points in input images'.format(self.METHOD_NAME))
        return keypoints, descriptors



from typing import Union
from collections import OrderedDict
from loguru import logger
import numpy as np
import cv2

from baseImage import Image
from .exceptions import BaseError, CreateExtractorError
from .settings import CVSTRATEGY, CVPARAMS
from .utils import print_run_time


class Findit(object):
    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True):
        """
        初始化
        :param threshold: 识别阈值(0~1)
        :param rgb: 是否使用rgb通道进行校验
        :return: None
        """
        self.threshold = threshold
        self.rgb = rgb
        self.match_methods = self.init_matching_methods()

    @print_run_time()
    def find_best_result(self, im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         threshold: Union[int, float] = None, rgb: bool = None):
        """
        使用CVSTRATEGY中的方法,依次运行。在im_source中,找到最符合im_search的范围坐标
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :param threshold: 识别阈值(0~1)
        :param rgb: 是否使用rgb通道进行校验
        :return: 如果找到了,则会返回一个Rect对象。没找到则会返回None
        """
        # 初始化参数
        threshold = threshold is None and self.threshold or threshold
        rgb = rgb is None and self.rgb or rgb

        for method_name, method in self.match_methods.items():
            if isinstance(method, list):
                for func in method:
                    result = self._try_find_best(func, im_source=im_source, im_search=im_search,
                                                 threshold=threshold, rgb=rgb)
                    if result:
                        return result
            else:
                result = self._try_find_best(method, im_source=im_source, im_search=im_search,
                                             threshold=threshold, rgb=rgb)
                if result:
                    return result

        return None

    def find_all_results(self, im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                         threshold: Union[int, float] = None, max_count: int = 10, rgb: bool = None):
        """
        通过特征点匹配,在im_source中,找到符合im_search的范围坐标集合
        :param im_source: 待匹配图像
        :param im_search: 图片模板
        :param max_count: 最多可以返回的匹配数量
        :param threshold: 识别阈值(0~1)
        :param rgb: 是否使用rgb通道进行校验
        :return: 如果找到了,则会返回一个Rect对象。没找到则会返回None
        """
        threshold = threshold is None and self.threshold or threshold
        rgb = rgb is None and self.rgb or rgb

        for method_name, method in self.match_methods.items():
            if isinstance(method, list):
                for func in method:
                    result = self._try_find_all(func, im_source=im_source, im_search=im_search,
                                                threshold=threshold, rgb=rgb, max_count=max_count)
                    if result:
                        return result
            else:
                result = self._try_find_all(method, im_source=im_source, im_search=im_search,
                                            threshold=threshold, rgb=rgb, max_count=max_count)
                if result:
                    return result

        return None

    @staticmethod
    def _try_find_best(func, im_source, im_search, threshold, rgb):
        """
        尝试运行find_best_result,返回找到的坐标
        """
        try:
            match_result = func.find_best_result(im_source=im_source, im_search=im_search, threshold=threshold, rgb=rgb)
        except BaseError as err:
            logger.error('{} param:{}', err, func.get_extractor_parameters())
            return None
        else:
            return match_result

    @staticmethod
    def _try_find_all(func, im_source, im_search, threshold, rgb, max_count):
        """
        尝试运行find_best_result,返回找到的坐标
        """
        try:
            match_result = func.find_all_results(im_source=im_source, im_search=im_search,
                                                 threshold=threshold, rgb=rgb, max_count=max_count)
        except BaseError as err:
            logger.error('{} param:{}', err, func.get_extractor_parameters())
            return None
        else:
            return match_result

    @staticmethod
    def init_matching_methods():
        MATCH_METHODS = OrderedDict()

        for method in CVSTRATEGY:
            if method.__name__ in CVPARAMS:
                param_config = CVPARAMS[method.__name__]
            else:
                param_config = None

            if isinstance(param_config, list):
                MATCH_METHODS[method.__name__] = []
                for cfg in param_config:
                    func = Findit._create_extractor(method, cfg)
                    if func:
                        MATCH_METHODS[method.__name__].append(func)
            elif isinstance(param_config, dict):
                func = Findit._create_extractor(method, param_config)
                if func:
                    MATCH_METHODS[method.__name__] = func
            else:
                func = Findit._create_extractor(method)
                if func:
                    MATCH_METHODS[method.__name__] = func

        return MATCH_METHODS

    @staticmethod
    def _create_extractor(method, config=None):
        try:
            if config:
                func = method(**config)
            else:
                func = method()
        except CreateExtractorError:
            logger.error('create {} params=({}) error', method.__name__, config)
            return None
        else:
            return func



# -*- coding: utf-8 -*-
import cv2
import numpy as np
from typing import Union, Tuple
from baseImage import Image

from image_registration.template_matching.matchTemplate import MatchTemplate
from image_registration.exceptions import NoModuleError


class CudaMatchTemplate(MatchTemplate):
    METHOD_NAME = "cuda_tpl"

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True):
        """
        初始化
        :param threshold: 识别阈值(0~1)
        :param rgb: 是否使用rgb通道进行校验
        :return: None
        """
        super(CudaMatchTemplate, self).__init__(threshold, rgb)
        try:
            self.matcher = cv2.cuda.createTemplateMatching(cv2.CV_8U, cv2.TM_CCOEFF_NORMED)
        except AttributeError:
            raise NoModuleError('create CUDA TemplateMatching Error')

    @staticmethod
    def check_detection_input(im_source: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes],
                              im_search: Union[Image, str, np.ndarray, cv2.cuda_GpuMat, bytes]) -> Tuple[Image, Image]:
        if not isinstance(im_source, Image):
            im_source = Image(im_source)
        if not isinstance(im_search, Image):
            im_search = Image(im_search)

        im_source.transform_gpu()
        im_search.transform_gpu()
        return im_source, im_search

    def _get_template_result_matrix(self, im_source: Image, im_search: Image):
        """求取模板匹配的结果矩阵."""
        s_gray, i_gray = im_search.rgb_2_gray(), im_source.rgb_2_gray()
        res = self.matcher.match(i_gray, s_gray)
        return res.download()

    def cuda_cal_rgb_confidence(self, img_src_rgb, img_sch_rgb):
        """
        计算两张图片图片rgb三通道的置信度
        :param img_src_rgb: 待匹配图像
        :param img_sch_rgb: 图片模板
        :return: 最小置信度
        """
        img_src_rgb, img_sch_rgb = img_src_rgb.download(), img_sch_rgb.download()
        # img_sch_rgb = cv2.cuda.copyMakeBorder(img_sch_rgb, 10, 10, 10, 10, cv2.BORDER_REPLICATE)
        # # 转HSV强化颜色的影响
        # img_src_rgb = cv2.cuda.cvtColor(img_src_rgb, cv2.COLOR_BGR2HSV)
        # img_sch_rgb = cv2.cuda.cvtColor(img_sch_rgb, cv2.COLOR_BGR2HSV)
        src_bgr, sch_bgr = cv2.cuda.split(img_src_rgb), cv2.cuda.split(img_sch_rgb)
        # 计算BGR三通道的confidence，存入bgr_confidence:
        bgr_confidence = [0, 0, 0]
        for i in range(3):
            res_temp = self.matcher.match(sch_bgr[i], src_bgr[i]).download()
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res_temp)
            bgr_confidence[i] = max_val
        return min(bgr_confidence)

    def _get_confidence_from_matrix(self, img_crop, im_search, max_val, rgb):
        """根据结果矩阵求出confidence."""
        # 求取可信度:
        if rgb:
            # 如果有颜色校验,对目标区域进行BGR三通道校验:
            # confidence = self.cuda_cal_rgb_confidence(img_crop, im_search)
            confidence = MatchTemplate.cal_rgb_confidence(img_crop, im_search)
        else:
            confidence = max_val
        return confidence



from setuptools import setup, find_packages

setup(
    name='py-image-registration',
    version='1.0.14',
    author='hakaboom',
    license="Apache License 2.0",
    author_email='1534225986@qq.com',
    description='image registration by ORB/SIFT/SURF/CUDA_SURF/ORB/CUDA_ORB/matchTemplate',
    url='https://github.com/hakaboom/py_image_registration',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['colorama>=0.4.4',
                      "loguru>=0.5.3",
                      "baseImage==1.0.6",
],
)


# -*- coding: utf-8 -*-
import cv2
from typing import Union
from image_registration.keypoint_matching.kaze import KAZE
from image_registration.exceptions import (CreateExtractorError)


class SURF(KAZE):
    METHOD_NAME = "SURF"

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True,
                 hessianThreshold: int = 400, nOctaves: int = 4, nOctaveLayers: int = 3,
                 extended: bool = True, upright: bool = False):
        super(SURF, self).__init__(threshold, rgb)
        # 初始化参数
        self.extractor_parameters = dict(
            hessianThreshold=hessianThreshold, nOctaves=nOctaves, nOctaveLayers=nOctaveLayers,
            extended=extended, upright=upright,
        )
        try:
            self.detector = cv2.xfeatures2d.SURF_create(**self.extractor_parameters)
        except Exception:
            raise CreateExtractorError('create surf extractor error')



# -*- coding: utf-8 -*-
from typing import Union
import cv2

from image_registration.keypoint_matching.kaze import KAZE
from image_registration.exceptions import (CreateExtractorError)


class AKAZE(KAZE):
    METHOD_NAME = "AKAZE"

    def __init__(self, threshold: Union[int, float] = 0.8, rgb: bool = True,
                 descriptor_type: int = cv2.AKAZE_DESCRIPTOR_MLDB,
                 descriptor_size: int = 0, descriptor_channels: int = 3,
                 _threshold: float = 0.001, nOctaves: int = 4, nOctaveLayers: int = 4,
                 diffusivity: int = cv2.KAZE_DIFF_PM_G2):
        super(AKAZE, self).__init__(threshold, rgb)
        self.extractor_parameters = dict(
            descriptor_type=descriptor_type, descriptor_size=descriptor_size, descriptor_channels=descriptor_channels,
            threshold=_threshold, diffusivity=diffusivity,
            nOctaves=nOctaves, nOctaveLayers=nOctaveLayers,
        )
        try:
            self.detector = cv2.AKAZE_create(**self.extractor_parameters)
        except Exception:
            raise CreateExtractorError('create akaze extractor error')

    def create_matcher(self) -> cv2.BFMatcher:
        matcher = cv2.BFMatcher_create(cv2.NORM_L1)
        return matcher
